//
//  Copyright 2011 Canonical Ltd.
//	
//  This program is free software: you can redistribute it and/or modify it
//  under the terms of the GNU Affero General Public License version 3,
//  as published by the Free Software Foundation.
//	
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranties of
//  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#import <AssetsLibrary/AssetsLibrary.h>
#import <MobileCoreServices/UTType.h>

#import "Reachability.h"

#import "U1AccountManager.h"
#import "U1Asset.h"
#import "U1AssetRepresenationDataProvider.h"
#import "U1AutoUploadOperation.h"
#import "U1DataRepository.h"
#import "U1AutoUploadsManager.h"
#import "U1FileNode.h"
#import "U1FilesClient.h"
#import "U1FolderNode.h"
#import "U1Volume.h"


@interface U1AutoUploadsManager ()
@property (retain) U1FilesClient *filesClient;
@property (retain) NSOperationQueue *autoUploadPreparationQueue;
@property (retain) NSManagedObjectID *autoUploadsFolderID;
- (void)walkAssetsInLibrary;
- (NSString *)generateFilenameForAsset:(ALAsset *)asset;
- (void)processAssetBatch:(NSArray*)batch inGroup:(ALAssetsGroup *)group toArray:(NSMutableArray*)ops;
- (void)autoUploadSettingsChanged:(NSNotification*)notification;
- (BOOL)isPending:(NSString *)filename;
- (NSString*)remoteUploadFolderPath;
- (void)fetchRemoteUploadFolderContents;
- (void)queuePreviousUploads;
- (void)volumesFetched:(NSNotification *)notification;
@end


@implementation U1AutoUploadsManager

@synthesize filesClient, dataRepository, assetsLibrary;
@synthesize autoUploadPreparationQueue, autoUploadsFolderID;
@synthesize remoteUploadFolder;

+ (U1AutoUploadsManager *)sharedAutoUploadsManager;
{
	static U1AutoUploadsManager *sharedAutoUploadsManager = nil;
	if (sharedAutoUploadsManager == nil)
	{
		sharedAutoUploadsManager = [[U1AutoUploadsManager alloc] init];
	}
	return sharedAutoUploadsManager;
}

- (id)init
{
    if (!(self = [super init]))
		return nil;

	NSNotificationCenter *defaultCenter = [NSNotificationCenter defaultCenter];
	
	filesClient = [U1FilesClient sharedFilesClient];
	autoUploadPreparationQueue = [[NSOperationQueue alloc] init];
	[self.autoUploadPreparationQueue setMaxConcurrentOperationCount:1];
    assetsLibrary = [[ALAssetsLibrary alloc] init];
	[defaultCenter addObserver:self selector:@selector(libraryChanged:) name:ALAssetsLibraryChangedNotification object:nil];
	[defaultCenter addObserver:self selector:@selector(autoUploadSettingsChanged:) name:NSUserDefaultsDidChangeNotification object:[NSUserDefaults standardUserDefaults]];
	dataRepository = [U1DataRepository sharedDataRepository];
	[defaultCenter addObserver:self selector:@selector(reachabilityChanged:) name:kReachabilityChangedNotification object:nil];
	[defaultCenter addObserver:self selector:@selector(volumesFetched:) name:@"volumesFetched" object:nil];
	[defaultCenter addObserver:self selector:@selector(didSignOut:) name:U1AccountManagerDidSignOut object:nil];
    return self;
}

- (void)dealloc
{
	[[NSNotificationCenter defaultCenter] removeObserver:self];
	[assetsLibrary release];
	[autoUploadPreparationQueue release];
	[remoteUploadFolder release];
	[autoUploadsFolderID release];
    [super dealloc];
}

- (void)checkForNewAssets;
{
	[self walkAssetsInLibrary];
}

- (void)libraryChanged:(NSNotification*)notification;
{
	[self walkAssetsInLibrary];
}

- (void)autoUploadSettingsChanged:(NSNotification *)notification;
{
	NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
	
	if ([defaults boolForKey:@"auto_upload"])
	{
		// Turned it on, start your engines
		[self.autoUploadPreparationQueue setSuspended:NO];
		// trigger the fetch volumes with a no-op completionBlock. XXX: Minorly skeezy.
		[self.filesClient volumesWithResultsBlock:^(U1Volume *rootVolume, NSFetchedResultsController *cloudFoldersResultsController, U1DataSourceType resultsDataSourceType) {} errorBlock:^(NSError *error) {}];
	}
	else
	{
		[self.autoUploadPreparationQueue setSuspended:YES];
	}
}

- (void)reachabilityChanged:(NSNotification*)notification;
{
	Reachability *reachability = [notification object];
	NetworkStatus status = [reachability currentReachabilityStatus];
	switch (status)
	{
		case ReachableViaWiFi:
			[self.autoUploadPreparationQueue setSuspended:NO];
			break;
		case ReachableViaWWAN:
		{
			BOOL shouldSuspend = [[NSUserDefaults standardUserDefaults] boolForKey:@"wifi_only"];
			[self.autoUploadPreparationQueue setSuspended:shouldSuspend];
			break;
		}
		case NotReachable:
			[self.autoUploadPreparationQueue setSuspended:YES];
			break;
	}
}

- (void)volumesFetched:(NSNotification *)notification;
{
	NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];

	if ([defaults boolForKey:@"auto_upload"])
	{
		NSArray *volumes = [notification object];
		BOOL found = NO;
		U1Volume *volume = nil;
		for (U1Volume *testVolume in volumes) {
			if ([testVolume.nodePath isEqualToString:[self remoteUploadFolderPath]])
			{
				found = YES;
				volume = testVolume;
				break;
			}
		}
		if (!found)
		{
			static BOOL creatingUploadVolume = NO;
			
			if (creatingUploadVolume)
				return;
			
			creatingUploadVolume = YES;
			
			[self.filesClient createVolumeAtPath:[self remoteUploadFolderPath]
								 completionBlock:^(U1Volume *volume, NSError *error) {
									 creatingUploadVolume = NO;
									 if (error == nil)
									 {
										 self.remoteUploadFolder = volume.rootFolder;
										 self.autoUploadsFolderID = [volume.rootFolder objectID];
										 [self fetchRemoteUploadFolderContents];
									 }
									 else
									 {
										 NSLog(@"Error ensuring upload folder: %@", error);
									 }
								 }];	
		}
		else
		{
			self.remoteUploadFolder = volume.rootFolder;
			self.autoUploadsFolderID = [volume.rootFolder objectID];
			[self fetchRemoteUploadFolderContents];
		}
	}
	// We received a list of volumes. If our upload folder is not among them, prime it.
}

- (void)didSignOut:(NSNotification*)notification;
{
	[self.autoUploadPreparationQueue cancelAllOperations];
}

- (void)fetchRemoteUploadFolderContents;
{
	// fetch the remote upload folder contents, and stash the children objects somewhere locally.
	[self.filesClient fetchFolderWithResourcePath:self.remoteUploadFolder.resourcePath
									 resultsBlock:^(U1Node *node, NSFetchedResultsController *childrenResultsController, U1DataSourceType resultsDataSourceType) {

										 if (resultsDataSourceType == U1DataSourceRemote)
										 {
											 [childrenResultsController performFetch:NULL];
											 // TODO: this
											 [self queuePreviousUploads];
											 [self checkForNewAssets];
										 }
									 }
									   errorBlock:^(NSError *error) {}];
}

- (void)walkAssetsInLibrary;
{
	static BOOL alreadyWalking = NO;
	
	if (alreadyWalking)
		return;
	
	alreadyWalking = YES;
	
	NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
	
	if ([defaults boolForKey:@"auto_upload"])
	{
		dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^{

			[self.assetsLibrary enumerateGroupsWithTypes:ALAssetsGroupSavedPhotos usingBlock:^(ALAssetsGroup *group, BOOL *stop) {
				
				NSMutableArray *ops = [NSMutableArray array];
				__block NSMutableArray *batch = [[NSMutableArray alloc] initWithCapacity:50];
				[group enumerateAssetsUsingBlock:^(ALAsset *asset, NSUInteger index, BOOL *stop) {
					
					if (asset == nil)
					{
						return;
					}
					
					[batch addObject:asset];
					
					if ([batch count] >= 50)
					{
						NSSortDescriptor *sortBy = [NSSortDescriptor sortDescriptorWithKey:@"defaultRepresentation.url.absoluteString" ascending:YES];
						[batch sortUsingDescriptors:[NSArray arrayWithObject:sortBy]];
						[self processAssetBatch:batch inGroup:group toArray:ops];
						[batch release];
						batch = [[NSMutableArray alloc] initWithCapacity:50];
					}
				}];
				if ([batch count] > 0) // Deal with last partial batch
				{
					NSSortDescriptor *sortBy = [NSSortDescriptor sortDescriptorWithKey:@"defaultRepresentation.url.absoluteString" ascending:YES];
					[batch sortUsingDescriptors:[NSArray arrayWithObject:sortBy]];
					[self processAssetBatch:batch inGroup:group toArray:ops];
				}
				[batch release];
				batch = nil;
				
				NSArray *sortedOps = [ops sortedArrayUsingDescriptors:[NSArray arrayWithObject:[NSSortDescriptor sortDescriptorWithKey:@"filename" ascending:YES]]];
				[sortedOps enumerateObjectsUsingBlock:^(U1AutoUploadOperation *operation, NSUInteger idx, BOOL *stop) {
					[self.autoUploadPreparationQueue addOperation:operation];
					[[NSNotificationCenter defaultCenter] postNotificationName:@"uploadQueued" object:nil];
				}];
				[ops removeAllObjects];
				alreadyWalking = NO;
			} failureBlock:^(NSError *error) {
				NSLog(@"Some error happened: %@", error);
			}];
			
		});
		
	}
	
}

- (void)processAssetBatch:(NSArray*)batch inGroup:(ALAssetsGroup *)group toArray:(NSMutableArray*)ops;
{
	NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
	
	// instead of dealing with a single asset, loop over our batch
	NSMutableArray *URLs = [batch valueForKeyPath:@"defaultRepresentation.url.absoluteString"];

	[self.dataRepository dispatchBlockWithManagedObjectContext:^(NSManagedObjectContext *context) {

		U1FolderNode *autoUploadsFolder = (id)[context objectWithID:self.autoUploadsFolderID];
		NSMutableSet *assetsToUpload = [NSMutableSet set];

		NSPredicate *p = [NSPredicate predicateWithFormat:@"url in %@ and (fileNode.parent = %@ OR fileNode = nil)", URLs, autoUploadsFolder];
		NSError *error = nil;
		NSSortDescriptor *sortBy = [NSSortDescriptor sortDescriptorWithKey:@"url" ascending:YES];
		NSArray *existingU1Assets = [self.dataRepository resultsForEntityClass:[U1Asset class]
																	  inContext:context
																  withPredicate:p
																sortDescriptors:[NSArray arrayWithObject:sortBy]
																		  error:&error];
		
		// If batch's asset isn't in existingU1Assets or existingU1Asset's generation is nil, we upload it.
		
		// walk batch, looking for 
		NSEnumerator *assetEnumerator = [existingU1Assets objectEnumerator];
		__block U1Asset *nextU1Asset = [assetEnumerator nextObject];
		[batch enumerateObjectsUsingBlock:^(ALAsset *asset, NSUInteger idx, BOOL *stop) {
			BOOL shouldUpload = NO;
			U1Asset *assetToUpload = nil;
			
			NSString *url = [[[asset defaultRepresentation] url] absoluteString];
			while (nextU1Asset && [nextU1Asset.url compare:url] == NSOrderedAscending)
				nextU1Asset = [assetEnumerator nextObject];
			
			if ([nextU1Asset.url isEqualToString:url])
			{
				// matched... if the generation is nil, still need to upload it, but don't need to create it
				if (nextU1Asset.generation == nil)
				{
					shouldUpload = YES;
					assetToUpload = nextU1Asset;
				}
			}
			else
			{
				// no U1asset, need to create/upload
				NSString *assetType = [asset valueForProperty:ALAssetPropertyType];
				BOOL canUpload = NO;
				if ([defaults boolForKey:@"include_video"])
				{
					canUpload = ([assetType isEqualToString:ALAssetTypePhoto] ||
								 [assetType isEqualToString:ALAssetTypeVideo]);
				}
				else
				{
					canUpload = [assetType isEqualToString:ALAssetTypePhoto];
				}
				
				if (canUpload) {
					shouldUpload = YES;
					assetToUpload = [self createU1AssetWithAsset:asset group:group URLString:url fileName:nil inFolder:autoUploadsFolder context:context];
				}
			}
			if (shouldUpload)
			{
				[assetsToUpload addObject:assetToUpload];
			}
		}];

		[context save:NULL];

		[assetsToUpload enumerateObjectsUsingBlock:^(U1Asset *u1asset, BOOL *stop) {
			
			NSOperation *op = [self queueAutoUploadForAsset:u1asset andRepresentation:nil];
			if (op)
			{
				[ops addObject:op];
			}
		}];
	}];

	
}

- (void)queuePreviousUploads;
{
	// We should poll our data repository for all U1Assets which have generation=nil. They *should* all have U1FileNodes associated with them. If the U1FileNode's parent equals our autoUploadsManager's remoteUploadFolder, then it's an auto-upload, and we let the LAM handle it. Otherwise, kick it off immediately to the FilesClient and get it back in the queue.
		NSPredicate *p = [NSPredicate predicateWithFormat:@"generation = %@", nil];
		NSError *error = nil;
		NSSortDescriptor *sortBy = [NSSortDescriptor sortDescriptorWithKey:@"filename" ascending:YES];
		NSArray *previouslyQueuedU1Assets = [self.dataRepository resultsForEntityClass:[U1Asset class]
																	 matchingPredicate:p
																   withSortDescriptors:[NSArray arrayWithObject:sortBy]
																				 error:&error];
		[previouslyQueuedU1Assets enumerateObjectsUsingBlock:^(U1Asset *u1asset, NSUInteger idx, BOOL *stop) {
			if (u1asset.fileNode.parent == self.remoteUploadFolder)
			{
				[self.assetsLibrary assetForURL:[NSURL URLWithString:u1asset.url]
									resultBlock:^(ALAsset *asset) {
										ALAssetRepresentation *rep = [asset defaultRepresentation];
										// Belongs to LAM, could punt and let it get picked up by checkForNewAssets
										NSOperation *op = [self queueAutoUploadForAsset:u1asset andRepresentation:rep];
										if (op)
											[self.autoUploadPreparationQueue addOperation:op];
									} failureBlock:^(NSError *error) {
										NSLog(@"Error: %@", error);
										// TODO: delete asset
									}];
			}
			else
			{
				NSURL *imageURL = [NSURL URLWithString:u1asset.url];
				
				[self.assetsLibrary assetForURL:imageURL
							   resultBlock:^(ALAsset *asset) {
								   ALAssetRepresentation *representation = [asset defaultRepresentation];
								   NSString *mimetype = (id)UTTypeCopyPreferredTagWithClass((CFStringRef)[representation UTI], kUTTagClassMIMEType);
								   
								   U1AssetRepresenationDataProvider *provider = [[U1AssetRepresenationDataProvider alloc] init];
								   provider.assetURL = imageURL;
								   
								   [self.filesClient uploadContentDataProvider:provider
																	  toFolder:u1asset.fileNode.parent
															  withResourceName:u1asset.filename
															   withContentType:mimetype
																  withPriority:NSOperationQueuePriorityVeryHigh
																 progressBlock:^(long long bytesUploaded, long long totalBytes) {}
															   completionBlock:^(U1FileNode *updatedNode, NSError *error) {
																   u1asset.generation = updatedNode.generation;
															   }];
							   } failureBlock:^(NSError *error) {
								   NSLog(@"Error: %@", error);
							   }];
			}
		}];
}

- (NSString *)remoteUploadFolderPath;
{
	return [NSString stringWithFormat:@"/~/Pictures - %@", [[UIDevice currentDevice] name]];
}

- (NSString *)generateFilenameForAsset:(ALAsset *)asset;
{
	// get the type of the asset
	NSString *filename = nil;
	ALAssetRepresentation *representation = [asset defaultRepresentation];
	
	if ([representation respondsToSelector:@selector(filename)]) // iOS 5+
	{
		filename = [representation filename];
	}
	else
	{
		NSString *ext = (id)UTTypeCopyPreferredTagWithClass((CFStringRef)[representation UTI], kUTTagClassFilenameExtension);
	
		int idx = (int)[self numberOfAssets];

		filename = [NSString stringWithFormat:@"IMG_%d.%@", idx, ext];
	}

	return filename;
}

- (U1Asset *)createU1AssetWithAsset:(ALAsset *)asset group:(ALAssetsGroup *)group URLString:(NSString *)urlString fileName:(NSString *)filename inFolder:(U1FolderNode *)folderNode context:(NSManagedObjectContext*)context;
{
	U1Asset *newAsset = [U1Asset insertInManagedObjectContext:context];
	newAsset.groupId = [group valueForProperty: ALAssetsGroupPropertyPersistentID];
	newAsset.url = urlString;
	if (filename == nil)
	{
		filename = [self generateFilenameForAsset:asset];
	}
	newAsset.filename = filename;
	NSString *resourcePath = [folderNode.resourcePath stringByAppendingPathComponent:filename];
	NSString *contentPath = [folderNode.contentPath stringByAppendingPathComponent:filename];
	
	__block BOOL found = NO;
	__block U1FileNode *node = nil;
	
	[folderNode.children enumerateObjectsUsingBlock:^(U1FileNode *childNode, BOOL *stop) {
		if ([childNode.name isEqualToString:filename]) {
			found = YES;
			newAsset.generation = childNode.generation;
			newAsset.fileNode = childNode;
			*stop = YES;
		}
	}];
	// Now we have an asset. We need to check to see if a corresponding FileNode is in our remoteUploadFolderChildren, and use it if it is.
	if (!found)
	{
		node = [U1FileNode insertInManagedObjectContext:context];
		node.resourcePath = resourcePath;
		node.kind = @"file";
		node.parent = folderNode;
		node.contentPath = contentPath;
		node.asset = newAsset;
	}
	
	return newAsset;
}

- (NSOperation*)queueAutoUploadForAsset:(U1Asset *)assetToUpload andRepresentation:(ALAssetRepresentation *)representation;
{
	// If the assetToUpload's generation is nil, but its fileNode's generation isn't, then the Node already exists on the server. Update the asset and move on
	if (assetToUpload.generation == nil && assetToUpload.fileNode.generation != nil)
	{
		assetToUpload.generation = assetToUpload.fileNode.generation;
		[self.dataRepository save:NULL];
	}
	
	if ((![self isPending:assetToUpload.filename]) && (assetToUpload.generation == nil))
	{
		U1AssetRepresenationDataProvider *provider = [[U1AssetRepresenationDataProvider alloc] init];
		provider.assetURL = [NSURL URLWithString:assetToUpload.url];
		
		// Let's create an operation!
		U1AutoUploadOperation *operation = [[U1AutoUploadOperation alloc] init];
		operation.dataProvider = provider;
		operation.assetID = [assetToUpload objectID];
		operation.folderID = [self.remoteUploadFolder objectID];
		operation.filename = assetToUpload.filename;
		// TODO:
//		operation.contentType = (id)UTTypeCopyPreferredTagWithClass((CFStringRef)[provider.representation UTI], kUTTagClassMIMEType);
		operation.priority = NSOperationQueuePriorityNormal;
		return [operation autorelease];
	}
	return nil;
}

- (int)numberOfAssets;
{
	NSError *error = nil;
	return [[self.dataRepository resultsForEntityClass:[U1Asset class]
									matchingPredicate:nil
								  withSortDescriptors:nil
												error:&error] count];
}

- (void)thumbnailForNode:(U1FileNode *)fileNode completionBlock:(void(^)(CGImageRef thumbnail))completionBlock;
{
	[self.assetsLibrary assetForURL:[NSURL URLWithString:fileNode.asset.url]
						resultBlock:^(ALAsset *asset) {
							completionBlock(asset.thumbnail);
						} failureBlock:^(NSError *error) {
							NSLog(@"Couldn't find the asset to get a thumbnail: %@", error);
						}];
}

- (BOOL)isPending:(NSString *)filename;
{
	__block BOOL isPending = NO;
	
	[[self.autoUploadPreparationQueue operations] enumerateObjectsUsingBlock:^(U1AutoUploadOperation *operation, NSUInteger idx, BOOL *stop) {
		if ([operation.filename isEqualToString:filename])
		{
			isPending = YES;
			*stop = YES;
		}
	}];
	
	return isPending;
}

- (int)pendingCount;
{
	__block int count = 0;
	
	[[self.autoUploadPreparationQueue operations] enumerateObjectsUsingBlock:^(U1AutoUploadOperation *operation, NSUInteger idx, BOOL *stop) {
		if (![operation isExecuting])
		{
			count++;
		}
	}];
	
	return count;
}

@end
