//
//  Copyright 2011 Canonical Ltd.
//	
//  This program is free software: you can redistribute it and/or modify it
//  under the terms of the GNU Affero General Public License version 3,
//  as published by the Free Software Foundation.
//	
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranties of
//  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#import "U1DataRepository.h"

#import <CoreData/CoreData.h>

#import "U1Node.h"

static U1DataRepository *sharedDataRepository = nil;

@interface U1DataRepository ()

@property (nonatomic, retain) NSManagedObjectModel *model;
@property (nonatomic, retain) NSManagedObjectContext *mainContext;
@property (nonatomic, retain) NSPersistentStoreCoordinator *storeCoordinator;

- (NSArray*)resultsForEntityClass:(Class)class inContext:(NSManagedObjectContext*)context withPredicate:(NSPredicate*)predicate sortDescriptors:(NSArray*)sortDescriptors error:(NSError**)error;

@end


@implementation U1DataRepository

@synthesize model, mainContext, storeCoordinator;

+ (U1DataRepository*)sharedDataRepository;
{
	if (sharedDataRepository == nil)
	{
		NSURL *cachesDirectory = [[[NSFileManager defaultManager] URLsForDirectory:NSCachesDirectory
																		 inDomains:NSUserDomainMask] lastObject];
		NSURL *storeURL = [NSURL URLWithString:@"U1Files.sqlite" relativeToURL:cachesDirectory];
		sharedDataRepository = [[self alloc] initWithStoreURL:storeURL];
	}
	return sharedDataRepository;
}

- (id)initWithStoreURL:(NSURL*)storeURL;
{
	if (!(self = [super init]))
		return nil;
	
	NSError *error = nil;
	
	NSURL *modelURL = [[NSBundle mainBundle] URLForResource:@"U1Files" withExtension:@"momd"];
	NSManagedObjectModel *mom = [[[NSManagedObjectModel alloc] initWithContentsOfURL:modelURL] autorelease];
	
	if (mom == nil)
		return nil;
	
	NSDictionary *options = [NSDictionary dictionaryWithObjectsAndKeys:
							 [NSNumber numberWithBool:YES], NSMigratePersistentStoresAutomaticallyOption,
							 [NSNumber numberWithBool:YES], NSInferMappingModelAutomaticallyOption,
							 nil];
	NSPersistentStoreCoordinator *coordinator = [[[NSPersistentStoreCoordinator alloc] initWithManagedObjectModel:mom] autorelease];
	if (![coordinator addPersistentStoreWithType:NSSQLiteStoreType configuration:nil URL:storeURL options:options error:&error])
	{
		NSLog(@"Error opening persistent store: %@", error);
		return nil;
	}
	
	NSManagedObjectContext *context = [[[NSManagedObjectContext alloc] init] autorelease];
	[context setPersistentStoreCoordinator:coordinator];
	[context setMergePolicy:NSOverwriteMergePolicy];
	
	self.model = mom;
	self.storeCoordinator = coordinator;
	self.mainContext = context;

	return self;
}

- (void)dealloc;
{
	[model release];
	[mainContext release];
	[storeCoordinator release];
	[super dealloc];
}

- (BOOL)save:(NSError**)error;
{
	return [self.mainContext save:error];
}

- (U1Node*)nodeWithResourcePath:(NSString *)resourcePath;
{
	NSParameterAssert(resourcePath);

	NSPredicate *p = [NSPredicate predicateWithFormat:@"resourcePath = %@", resourcePath];
	NSArray *nodes = [self resultsForEntityClass:[U1Node class] inContext:self.mainContext withPredicate:p sortDescriptors:nil error:NULL];
	U1Node *node = [nodes lastObject];
	return node;
}

- (NSArray*)resultsForEntityClass:(Class)entityClass matchingPredicate:(NSPredicate*)predicate withSortDescriptors:(NSArray*)sortDescriptors error:(NSError**)error;
{
	return [self resultsForEntityClass:entityClass inContext:self.mainContext withPredicate:predicate sortDescriptors:sortDescriptors error:error];
}

- (NSArray*)resultsForEntityClass:(Class)class inContext:(NSManagedObjectContext*)context withPredicate:(NSPredicate*)predicate sortDescriptors:(NSArray*)sortDescriptors error:(NSError**)error;
{
	NSParameterAssert(class != nil);
	NSParameterAssert(context != nil);

	NSFetchRequest *fetch = [[[NSFetchRequest alloc] init] autorelease];
	[fetch setEntity:[class entityInManagedObjectContext:context]];
	[fetch setPredicate:predicate];
	[fetch setSortDescriptors:sortDescriptors];
	NSArray *results = [context executeFetchRequest:fetch error:error];
	return results;
}

- (void)dispatchBlockWithMainContext:(void(^)(NSManagedObjectContext *context))block;
{
	NSParameterAssert(block != NULL);
	if (![NSThread isMainThread])
	{
		typeof(block) contextBlock = [block copy];
		dispatch_sync(dispatch_get_main_queue(), ^(void) {
			contextBlock(self.mainContext);
			[contextBlock release];
		});
	}
	else
	{
		block(self.mainContext);
	}
}

- (void)dispatchBlockWithManagedObjectContext:(void(^)(NSManagedObjectContext *context))block;
{
	NSParameterAssert(block != NULL);
	
	NSManagedObjectContext *context = [[NSManagedObjectContext alloc] init];
	id observation = [[[NSNotificationCenter defaultCenter] addObserverForName:NSManagedObjectContextDidSaveNotification object:context queue:nil usingBlock:^(NSNotification *note) {
		dispatch_async(dispatch_get_main_queue(), ^{
			[self.mainContext mergeChangesFromContextDidSaveNotification:note];
		});
	}] retain];
	[context setPersistentStoreCoordinator:self.storeCoordinator];
	block(context);
	[context release];
	[observation release];
}

- (void)dispatchAsyncBlockWithManagedObjectContext:(void(^)(NSManagedObjectContext *context))block;
{
	NSParameterAssert(block != NULL);
	
	typeof(block) contextBlock = [block copy];
	dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^(void) {
		NSManagedObjectContext *context = [[NSManagedObjectContext alloc] init];
		id observation = [[[NSNotificationCenter defaultCenter] addObserverForName:NSManagedObjectContextDidSaveNotification object:context queue:nil usingBlock:^(NSNotification *note) {
			dispatch_async(dispatch_get_main_queue(), ^{
				[self.mainContext mergeChangesFromContextDidSaveNotification:note];
			});
		}] retain];
		[context setPersistentStoreCoordinator:self.storeCoordinator];
		contextBlock(context);
		[contextBlock release];
		[context release];
		[observation release];
	});
}

@end
