//
//  Copyright 2011 Canonical Ltd.
//	
//  This program is free software: you can redistribute it and/or modify it
//  under the terms of the GNU Affero General Public License version 3,
//  as published by the Free Software Foundation.
//	
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranties of
//  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#import "U1FolderViewController.h"

#import <AssetsLibrary/AssetsLibrary.h>
#import <MobileCoreServices/UTType.h>
#import <MobileCoreServices/UTCoreTypes.h>

#import "U1Asset.h"
#import "U1AssetRepresenationDataProvider.h"
#import "U1ByteSizeValueTransformer.h"
#import "U1DataRepository.h"
#import "U1FileNode.h"
#import "U1FilesClient.h"
#import "U1FolderItemCell.h"
#import "U1FolderNode.h"
#import "U1LocalFileInfo.h"
#import "U1AutoUploadsManager.h"
#import "U1Node.h"
#import "UIActionSheet+U1Additions.h"


@interface U1FolderViewController () <UINavigationControllerDelegate, UIImagePickerControllerDelegate, NSFetchedResultsControllerDelegate>
@property (copy) NSString *resourcePath;
@property (retain) U1ByteSizeValueTransformer *byteSizeTransformer;
@property (retain) NSIndexPath *pendingIndexPath;
@property (retain) U1FolderNode *node;
@property (retain) NSMutableArray *nodes;
@property (retain) NSFetchedResultsController *resultsController;
- (void)loadInfoForNode;
- (void)deleteNodeAtIndexPath:(NSIndexPath*)indexPath;
- (void)configureCell:(U1FolderItemCell*)cell forNode:(U1Node*)aNode withInfo:(U1LocalFileInfo*)fileInfo;
@end


@implementation U1FolderViewController

@synthesize loadingCell;
@synthesize delegate, filesClient, node, resourcePath;
@synthesize byteSizeTransformer, pendingIndexPath, nodes, resultsController;

- (id)initWithNodeResourcePath:(NSString*)theResourcePath;
{
	self = [super initWithNibName:@"U1FolderViewController" bundle:nil];
	if (!self)
		return nil;
	self.title = [theResourcePath lastPathComponent];
	self.resourcePath = theResourcePath;
	self.byteSizeTransformer = [[U1ByteSizeValueTransformer new] autorelease];
	filesClient = [U1FilesClient sharedFilesClient];
	
	UIBarButtonItem *uploadItem = [[UIBarButtonItem alloc] initWithBarButtonSystemItem:UIBarButtonSystemItemAdd target:self action:@selector(uploadImage:)];
	[self.navigationItem setRightBarButtonItem:uploadItem];
	
	UILabel *titleLabel = [[UILabel alloc] initWithFrame:CGRectMake(0, 0, 100, 20)];
	// TODO: Ubuntu font
	[titleLabel setFont:[UIFont boldSystemFontOfSize:18.f]];
	[titleLabel setTextColor:[UIColor blackColor]];
	[titleLabel setShadowColor:[UIColor whiteColor]];
	[titleLabel setShadowOffset:CGSizeMake(-1, -1)];
	[titleLabel setText:self.title];
	[titleLabel sizeToFit];
	[titleLabel setOpaque:NO];
	[titleLabel setBackgroundColor:[UIColor clearColor]];
	[self.navigationItem setTitleView:titleLabel];
	[titleLabel release];
	[uploadItem release];
	
	return self;
}

- (void)dealloc;
{
	[[NSNotificationCenter defaultCenter] removeObserver:self name:U1FilesClientNodeStatusChangedNotification object:nil];
	[loadingCell release];
	[resourcePath release];
	[node release];
	[byteSizeTransformer release];
	[pendingIndexPath release];
	[nodes release];
	[resultsController release];
	[super dealloc];
}

- (void)viewDidLoad;
{
	[super viewDidLoad];
	[self refresh];
	
	UILongPressGestureRecognizer *longPress = [[UILongPressGestureRecognizer alloc] initWithTarget:self action:@selector(longPress:)];
	[longPress setMinimumPressDuration:1.f];
	[self.tableView addGestureRecognizer:longPress];
	[longPress release];
}

- (void)viewDidUnload;
{
	[[NSNotificationCenter defaultCenter] removeObserver:self name:U1FilesClientNodeStatusChangedNotification object:nil];
	[loadingCell release], loadingCell = nil;
	[super viewDidUnload];
}

- (void)viewDidAppear:(BOOL)animated;
{
	[super viewDidAppear:animated];
	[self.tableView deselectRowAtIndexPath:[self.tableView indexPathForSelectedRow] animated:YES];
	
	[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(nodeStatusChanged:) name:U1FilesClientNodeStatusChangedNotification object:nil];
}

- (void)viewWillDisappear:(BOOL)animated;
{
	[[NSNotificationCenter defaultCenter] removeObserver:self name:U1FilesClientNodeStatusChangedNotification object:nil];
}

- (void)nodeStatusChanged:(NSNotification*)notification;
{
	U1Node *aNode = [[notification userInfo] objectForKey:@"node"];
	U1LocalFileInfo *fileInfo = [[notification userInfo] objectForKey:@"fileInfo"];
	
	if (aNode.parent != self.node)
		return;
	
	NSIndexPath *indexPath = [self.resultsController indexPathForObject:aNode];
	U1FolderItemCell *cell = (id)[self.tableView cellForRowAtIndexPath:indexPath];
	if (cell)
	{
		[self configureCell:cell forNode:aNode withInfo:fileInfo];
	}
}

- (void)longPress:(UILongPressGestureRecognizer*)longPress;
{
	if (longPress.state == UIGestureRecognizerStateBegan)
	{
		CGPoint p = [longPress locationInView:self.tableView];
		NSIndexPath *indexPath = [self.tableView indexPathForRowAtPoint:p];
		U1Node *childNode = [self.nodes objectAtIndex:indexPath.row];

		UIActionSheet *actionSheet = [UIActionSheet new];
		
		[actionSheet setTitle:[[childNode path] lastPathComponent]];
		[actionSheet setDestructiveButtonIndex:[actionSheet addButtonWithTitle:@"Delete"]];
		NSInteger renameButtonIndex = [actionSheet addButtonWithTitle:@"Rename"];
		NSInteger publishButtonIndex = -1;
		NSInteger copyURLButtonIndex = -1;
		if ([childNode isFile])
		{
			U1FileNode *fileNode = (id)childNode;
			if ([[fileNode isPublic] boolValue])
			{
				publishButtonIndex = [actionSheet addButtonWithTitle:@"Unpublish"];
				copyURLButtonIndex = [actionSheet addButtonWithTitle:@"Copy Public URL"];
			}
			else
			{
				publishButtonIndex = [actionSheet addButtonWithTitle:@"Publish"];
			}
		}
		[actionSheet setCancelButtonIndex:[actionSheet addButtonWithTitle:@"Cancel"]];

		self.pendingIndexPath = indexPath;
		[actionSheet showInView:self.view withCompletionBlock:^(NSInteger buttonIndex) {
			if (buttonIndex == [actionSheet destructiveButtonIndex])
			{
				[self deleteNodeAtIndexPath:self.pendingIndexPath];
			}
			else if (buttonIndex == renameButtonIndex)
			{
				NSLog(@"rename!"); // TODO:
			}
			else if (buttonIndex == publishButtonIndex)
			{
				U1FileNode *fileNode = (id)childNode;
				if ([[fileNode isPublic] boolValue])
					[self.filesClient unpublishNode:fileNode completionBlock:^(U1FileNode *updatedNode, NSError *error) {
					}];
				else
					[self.filesClient publishNode:fileNode completionBlock:^(U1FileNode *updatedNode, NSError *error) {
					}];
			}
			else if (buttonIndex == copyURLButtonIndex)
			{
				U1FileNode *fileNode = (id)childNode;
				NSArray *types = UIPasteboardTypeListString;
				[[UIPasteboard generalPasteboard] setValue:[fileNode publicURLString] forPasteboardType:[types lastObject]];
			}
			else if (buttonIndex == [actionSheet cancelButtonIndex])
			{
//				NSLog(@"cancel");
			}
			self.pendingIndexPath = nil;
		}];
	}
}

- (void)uploadImage:(id)sender;
{
	UIImagePickerController *imagePicker = [[UIImagePickerController alloc] init];
	imagePicker.delegate = self;
	[self presentModalViewController:imagePicker animated:YES];
}

- (void)imagePickerController:(UIImagePickerController *)picker didFinishPickingMediaWithInfo:(NSDictionary *)info;
{
	// TODO: check for jpg/png
	// TODO: handle videos (stream)
	
	[picker dismissModalViewControllerAnimated:YES];
	
	NSURL *imageURL = [info objectForKey:UIImagePickerControllerReferenceURL];
	U1AutoUploadsManager *autoUploadsManager = [U1AutoUploadsManager sharedAutoUploadsManager];
	ALAssetsLibrary *assetsLibrary = autoUploadsManager.assetsLibrary;
	
	[assetsLibrary assetForURL:imageURL
				   resultBlock:^(ALAsset *asset) {
					   ALAssetRepresentation *representation = [asset defaultRepresentation];
					   CFStringRef uti = (CFStringRef)[representation UTI];
					   NSString *mimetype = (id)UTTypeCopyPreferredTagWithClass(uti, kUTTagClassMIMEType);
					   NSString *assetType = nil;
					   if (UTTypeConformsTo(uti, kUTTypeImage))
					   {
						   assetType = @"Image";
					   }
					   else if (UTTypeConformsTo(uti, kUTTypeAudiovisualContent))
					   {
						   assetType = @"Video";
					   }
					   else
					   {
						   assetType = @"File"; // XXX: Extend to something better when we bring in external files
					   }
					   NSDate *assetDate = [asset valueForProperty:ALAssetPropertyDate];

					   NSDateFormatter *dateFormatter = [[NSDateFormatter alloc] init];
					   [dateFormatter setDateStyle:NSDateFormatterMediumStyle];
					   [dateFormatter setTimeStyle:NSDateFormatterMediumStyle];
					   
					   NSString *ext = (id)UTTypeCopyPreferredTagWithClass(uti, kUTTagClassFilenameExtension);
					   NSString *filename = [NSString stringWithFormat:@"%@ %@.%@", assetType, [dateFormatter stringFromDate:assetDate], ext];
					   [dateFormatter release];
					   
					   // TODO: fix!
					   U1Asset *u1asset = [autoUploadsManager createU1AssetWithAsset:asset group:nil URLString:[imageURL absoluteString] fileName:filename inFolder:self.node context:nil];
					   
					   if (u1asset.generation == nil) // Prevent duplicate uploading to the same folder
					   {
						   U1AssetRepresenationDataProvider *provider = [[U1AssetRepresenationDataProvider alloc] init];
						   provider.assetURL = [NSURL URLWithString:[u1asset url]];
						   dispatch_async(dispatch_get_main_queue(), ^(void) {
							   [self.filesClient uploadContentDataProvider:provider
																  toFolder:self.node
														  withResourceName:filename
														   withContentType:mimetype
															  withPriority:NSOperationQueuePriorityVeryHigh
															 progressBlock:^(long long bytesUploaded, long long totalBytes) {}
														   completionBlock:^(U1FileNode *updatedNode, NSError *error) {
															   
															   if (error)
															   {
																   // TODO: remove direct access to data repository
																   [[U1DataRepository sharedDataRepository] dispatchBlockWithMainContext:^(NSManagedObjectContext *context) {
																	   [context deleteObject:u1asset.fileNode];
																   }];
															   }

															   u1asset.generation = updatedNode.generation;
														   }];
						   });
					   }
					   
					   [[U1DataRepository sharedDataRepository] save:NULL];

				   } failureBlock:^(NSError *error) {
					   NSLog(@"Error: %@", error);
				   }];
	
}

- (void)configureCell:(U1FolderItemCell*)cell forNode:(U1Node*)aNode withInfo:(U1LocalFileInfo*)fileInfo;
{
	cell.nameLabel.text = [aNode.resourcePath lastPathComponent];
	
	BOOL hasChildren = [[aNode hasChildren] boolValue];
	[cell setAccessoryType:(hasChildren ? UITableViewCellAccessoryDisclosureIndicator : UITableViewCellAccessoryNone)];
	
	if (aNode.generation == nil)
	{
		[cell.uploadIndicator setHidden:NO];
		[cell.uploadIndicator startAnimating];
	}
	else
	{
		[cell.uploadIndicator setHidden:YES];
		[cell.uploadIndicator stopAnimating];
	}
	
	[cell.imageView setImage:nil];
	if ([aNode isFolder])
	{
		[cell.imageView setImage:[UIImage imageNamed:@"ic_folder"]];
		[cell.detailLabel setText:[aNode.modificationDate description]];
	}
	else
	{
		[cell.uploadProgressView setProgress:fileInfo.uploadPercentage];
		[cell.uploadProgressView setHidden:!fileInfo.isUploading];
		[cell.detailLabel setHidden:fileInfo.isUploading];

		NSString *contentPath = aNode.contentPath;
		NSURL *contentURL = [NSURL fileURLWithPath:contentPath];
		UIDocumentInteractionController *docController = [UIDocumentInteractionController interactionControllerWithURL:contentURL];
		[cell.imageView setImage:[docController.icons lastObject]];
		NSNumber *size = [aNode valueForKey:@"size"];
		NSString *sizeString = [self.byteSizeTransformer transformedValue:size];
		NSString *detailString = sizeString;
		if (aNode.modificationDate)
			detailString = [sizeString stringByAppendingFormat:@", modified %@", aNode.modificationDate];
		[cell.detailLabel setText:detailString];
	}
}


#pragma mark UITableViewDataSource Methods

- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section;
{
	if (!self.resultsController)
		return 1;
	
	return [[[self.resultsController sections] objectAtIndex:section] numberOfObjects];
}

- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath;
{
	if (!self.resultsController)
	{
		return self.loadingCell;
	}
	
	U1FolderItemCell *cell = [U1FolderItemCell cellForTableView:tableView];
	
	U1Node *childNode = [self.resultsController objectAtIndexPath:indexPath];
	

	[self configureCell:cell forNode:childNode withInfo:nil];
	
	return cell;
}


#pragma mark UITableViewDelegate Methods

- (void)tableView:(UITableView *)tableView didSelectRowAtIndexPath:(NSIndexPath *)indexPath;
{
	U1Node *childNode = [self.resultsController objectAtIndexPath:indexPath];
	[self.delegate folderController:self didSelectNode:childNode];
}

- (void)tableView:(UITableView *)tableView commitEditingStyle:(UITableViewCellEditingStyle)editingStyle forRowAtIndexPath:(NSIndexPath *)indexPath;
{
	if (editingStyle != UITableViewCellEditingStyleDelete)
		return;
	
	[self deleteNodeAtIndexPath:indexPath];
}


- (void)refresh;
{
	[self loadInfoForNode];
}

- (UIView *)tableView:(UITableView *)tableView viewForFooterInSection:(NSInteger)section;
{
	UILabel *footerLabel = [[[UILabel alloc] initWithFrame:CGRectMake(0, 0, 320, 22)] autorelease];
	footerLabel.backgroundColor = [[UIColor lightGrayColor] colorWithAlphaComponent:0.85f];
	[footerLabel setFont:[UIFont systemFontOfSize:14.f]];
	[footerLabel setShadowColor:[UIColor whiteColor]];
	[footerLabel setShadowOffset:CGSizeMake(0, 1)];
	[footerLabel setTextAlignment:UITextAlignmentCenter];

	int numItems = [[[self.resultsController sections] objectAtIndex:0] numberOfObjects];
	NSNumber *folderRawSize = [[self.resultsController fetchedObjects] valueForKeyPath:@"@sum.size"];
	NSString *folderSize = [self.byteSizeTransformer transformedValue:folderRawSize];
	
	[footerLabel setText:[NSString stringWithFormat:@"%d items, %@", numItems, folderSize]];

	return footerLabel;
}

- (CGFloat)tableView:(UITableView *)tableView heightForFooterInSection:(NSInteger)section;
{
	return 22.0f;
}

#pragma mark Private Methods

- (void)loadInfoForNode;
{
	NSParameterAssert(self.resourcePath != nil);

	[self.filesClient fetchFolderWithResourcePath:self.resourcePath resultsBlock:^(U1Node *theNode, NSFetchedResultsController *childrenResultsController, U1DataSourceType resultsDataSourceType) {
		
		self.node = (id)theNode;
		if (childrenResultsController != self.resultsController)
		{
			self.resultsController = childrenResultsController;
			[self.resultsController setDelegate:self];
			[self.tableView reloadData];
		}

		if (resultsDataSourceType == U1DataSourceRemote)
			[self stopLoading];
		
	} errorBlock:^(NSError *error) {
		NSLog(@"error: %@", error);
		[self stopLoading];
	}];
}

- (void)deleteNodeAtIndexPath:(NSIndexPath*)indexPath;
{
	U1Node *childNode = [self.resultsController objectAtIndexPath:indexPath];
	[self.filesClient deleteNode:childNode completionBlock:^(NSError *error) {
		if (error)
		{
			// TODO:
			return;
		}
	}];
}


#pragma mark NSFetchedResultsControllerDelegate Methods


- (void)controllerWillChangeContent:(NSFetchedResultsController *)controller;
{
	[self.tableView beginUpdates];
}


- (void)controller:(NSFetchedResultsController *)controller didChangeSection:(id <NSFetchedResultsSectionInfo>)sectionInfo
		   atIndex:(NSUInteger)sectionIndex forChangeType:(NSFetchedResultsChangeType)type;
{
	switch (type)
	{
		case NSFetchedResultsChangeInsert:
			[self.tableView insertSections:[NSIndexSet indexSetWithIndex:sectionIndex]
						  withRowAnimation:UITableViewRowAnimationFade];
			break;
			
		case NSFetchedResultsChangeDelete:
			[self.tableView deleteSections:[NSIndexSet indexSetWithIndex:sectionIndex]
						  withRowAnimation:UITableViewRowAnimationFade];
			break;
	}
}

- (void)controller:(NSFetchedResultsController *)controller didChangeObject:(id)anObject
	   atIndexPath:(NSIndexPath *)indexPath forChangeType:(NSFetchedResultsChangeType)type
	  newIndexPath:(NSIndexPath *)newIndexPath;
{
	
	UITableView *tableView = self.tableView;
	
	switch (type)
	{
			
		case NSFetchedResultsChangeInsert:
			[tableView insertRowsAtIndexPaths:[NSArray arrayWithObject:newIndexPath]
							 withRowAnimation:UITableViewRowAnimationMiddle];
			break;
			
		case NSFetchedResultsChangeDelete:
			[tableView deleteRowsAtIndexPaths:[NSArray arrayWithObject:indexPath]
							 withRowAnimation:UITableViewRowAnimationTop];
			break;
			
		case NSFetchedResultsChangeUpdate:
			// TODO: 
			//            [self configureCell:[tableView cellForRowAtIndexPath:indexPath]
			//					atIndexPath:indexPath];
			[self.tableView reloadRowsAtIndexPaths:[NSArray arrayWithObject:indexPath]
								  withRowAnimation:UITableViewRowAnimationNone];
			break;
			
		case NSFetchedResultsChangeMove:
			[tableView deleteRowsAtIndexPaths:[NSArray arrayWithObject:indexPath]
							 withRowAnimation:UITableViewRowAnimationFade];
			[tableView insertRowsAtIndexPaths:[NSArray arrayWithObject:newIndexPath]
							 withRowAnimation:UITableViewRowAnimationFade];
			break;
	}
}

- (void)controllerDidChangeContent:(NSFetchedResultsController *)controller;
{
	[self.tableView endUpdates];
}

@end
