//
//  Copyright 2011 Canonical Ltd.
//	
//  This program is free software: you can redistribute it and/or modify it
//  under the terms of the GNU Affero General Public License version 3,
//  as published by the Free Software Foundation.
//	
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranties of
//  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#import "U1LoginController.h"

#import "JSONKit.h"
#import "OAuthConsumer.h"
#import "NSString+Extras.h"

#import "U1AccountManager.h"
#import "UOHTTPFetchOperation.h"
#import "UOJSONFetchOperation.h"
#import "UOSSOCredentialsViewController.h"


static NSString *const UOSingleSignOnBaseURL = @"https://login.ubuntu.com/api/1.0/";

static NSString *UOSSOEncodedStringFromParametersDictionary(NSDictionary *arguments);


@interface U1LoginController ()
@property (retain) UIViewController *loginViewController;
@property (retain) NSOperationQueue *operationQueue;
- (void)loginWithUsername:(NSString*)username password:(NSString*)password;
- (void)handleError:(NSError*)error;
@end


@implementation U1LoginController

@synthesize delegate, loginViewController, operationQueue;

- (id)init;
{
	if (!(self = [super init]))
		return nil;
	
	UOSSOCredentialsViewController *rootController = [[[UOSSOCredentialsViewController alloc] initWithNibName:@"UOSSOCredentialsViewController" bundle:nil] autorelease];
	rootController.title = NSLocalizedString(@"Log In to Ubuntu One", @"");
	
	__block id this = self;
	rootController.loginAction = ^ (NSString *username, NSString *password)
	{
		[this loginWithUsername:username password:password];
	};
	
	UINavigationController *navController = [[UINavigationController alloc] initWithRootViewController:rootController];
//	navController.navigationBar.barStyle = UIBarStyleBlack;
//	navController.navigationBarHidden = YES;
	loginViewController = navController;
	
	operationQueue = [[NSOperationQueue alloc] init];
	
	return self;
}

- (void)dealloc
{
	delegate = nil;
	[loginViewController release];
	[operationQueue release];
    [super dealloc];
}

- (void)loginWithUsername:(NSString*)username password:(NSString*)password;
{
	// - show login spinner
	[self.loginViewController.view setUserInteractionEnabled:NO];
	
	// 1. get/create U1 oauth token
	NSString *tokenName = [NSString stringWithFormat:@"Ubuntu One @ %@", [[UIDevice currentDevice] name]];
	NSDictionary *params = [NSDictionary dictionaryWithObjectsAndKeys:
							@"authenticate", @"ws.op",
							tokenName, @"token_name",
							nil];
	NSString * paramsString = UOSSOEncodedStringFromParametersDictionary(params);
	
	NSURL *url = [NSURL URLWithString:[NSString stringWithFormat:@"%@authentications?%@", UOSingleSignOnBaseURL, paramsString]];
	NSMutableURLRequest *request = [NSMutableURLRequest requestWithURL:url];
	[request setHTTPMethod:@"GET"];
	
	NSURLCredential *credential = [NSURLCredential credentialWithUser:username password:password persistence:NSURLCredentialPersistenceNone];
	UOJSONFetchOperation *jsonOperation = [UOJSONFetchOperation jsonOperationWithRequest:request action:^(NSObject *jsonResult, NSError *error) {
		
		// 2. import the new token into Ubuntu One
		NSDictionary *oauthResponse = (id)jsonResult;
		
		if (error)
		{
			[self handleError:error];
			return;
		}
		
		NSString *encodedUsername = [username urlParameterEncodedString];
		NSURL *u1TokenURL = [NSURL URLWithString:[NSString stringWithFormat:@"https://one.ubuntu.com/oauth/sso-finished-so-get-tokens/%@", encodedUsername]];
		
		OAConsumer *consumer = [[[OAConsumer alloc] initWithKey:[oauthResponse objectForKey:@"consumer_key"] secret:[oauthResponse objectForKey:@"consumer_secret"]] autorelease];
		OAToken *token = [[[OAToken alloc] initWithKey:[oauthResponse objectForKey:@"token"] secret:[oauthResponse objectForKey:@"token_secret"]] autorelease];

		[[U1AccountManager sharedAccountManager] storeConsumerKey:consumer.key consumerSecret:consumer.secret token:token.key tokenSecret:token.secret];
		
		OAPlaintextSignatureProvider *signatureProvider = [[[OAPlaintextSignatureProvider alloc] init] autorelease];
		OAMutableURLRequest *oauthRequest = [[[OAMutableURLRequest alloc] initWithURL:u1TokenURL consumer:consumer token:token realm:nil signatureProvider:signatureProvider] autorelease];
		[oauthRequest prepare];
		UOHTTPFetchOperation *tokenOperation = [UOHTTPFetchOperation httpOperationWithRequest:oauthRequest action:^(NSString *responseBody, NSError *error) {
			
			// 3. Get the phone creds using the new token
			
			if (error)
			{
				[self handleError:error];
				return;
			}
			
			[self.delegate loginControllerDidLogin];
			[self.loginViewController.view setUserInteractionEnabled:YES];
		
		}];
		[self.operationQueue addOperation:tokenOperation];
		
	}];
	jsonOperation.credential = credential;
	[self.operationQueue addOperation:jsonOperation];
}

- (void)handleError:(NSError*)error;
{
	NSString *errorMessage = [error localizedDescription];
	if ([error code] == 401)
		errorMessage = NSLocalizedString(@"Incorrect username or password.", nil);
	else if (errorMessage == nil)
		errorMessage = NSLocalizedString(@"An error occured trying to sign in, please try again.", @"Login Error Message");

	UIAlertView *alert = [[UIAlertView alloc] initWithTitle:NSLocalizedString(@"Ubuntu One", nil) message:errorMessage delegate:self cancelButtonTitle:@"Dismiss" otherButtonTitles:nil];
	[alert show];
	[alert release];
	[self.loginViewController.view setUserInteractionEnabled:YES];
	UOSSOCredentialsViewController *loginController = (id)[(UINavigationController*)[self loginViewController] topViewController];
	[loginController reset];
}

@end


static NSString *UOSSOEncodedStringFromParametersDictionary(NSDictionary *arguments)
{
	NSMutableArray *encodedArguments = [NSMutableArray array];
	[arguments enumerateKeysAndObjectsUsingBlock:^(id key, id value, BOOL *stop) {
		if (![key isEqualToString:@"ws.op"])
			value = [value JSONString];
		value = [value urlParameterEncodedString];
		NSString *argumentPair = [key stringByAppendingFormat:@"=%@", value];
		[encodedArguments addObject:argumentPair];
	}];
	
	return [encodedArguments componentsJoinedByString:@"&"];
}
